import torch
from torch import Tensor
from pathlib import Path
from typing import Tuple, List, Dict, Literal, Generator

from epilepsy_bids_loader import Segment


def read_json(json_file: str | Path) -> Dict:
    import json
    with open(json_file, "r") as f:
        meta = json.load(f)
    return meta


def read_yaml(path: Path) -> Dict:
    import yaml
    with open(path) as f:
        out: dict = yaml.safe_load(f)
        return out


def batch_segments(
    segments: List[Segment],
    batch_size: int,
    usage: Literal['all', 'train', 'test']
) -> Generator[Tuple[Tensor, Tensor, Tensor], None, None]:
    """
    Creates batches of data from a list of segments.

    Args:
        segments (List[Segment]): List of segments.
        batch_size (int): Size of the batch dimension "N".

    Returns:
        Generator[Tuple[Tensor, Tensor, Tensor]]: Tuple of:
        - x: (N, C, L)
        - y: (N,)
        - t: (N,)
    """
    # batches: List[Tuple[ndarray, ndarray, ndarray]] = []
    for i in range(0, len(segments), batch_size):
        j = i + batch_size
        batch: List[Segment] = segments[i : j]
        batch_x = []
        batch_y = []
        batch_t = []

        for seg in batch:
            x, y, t = seg.data(usage=usage)
            batch_x.append(x) # (1, C, L)
            batch_y.append(torch.tensor(1 if y.any() else 0)) # (L,) -> 1
            batch_t.append(t[0]) # (L,) -> 1, use the start of the seg
            continue

        batch = (
            torch.concatenate(batch_x, axis=0), # (N, C, L)
            torch.stack(batch_y, axis=0), # (N,)
            torch.stack(batch_t, axis=0) # (N,)
        )
        yield batch
    return



from numpy import ndarray
from torch import Tensor
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from collections import OrderedDict
from typing import Tuple, List, Any, Literal, Callable, Generator
from epilepsy_bids_loader.utils import Status


@dataclass
class Segment:
    smin: int
    smax: int
    subject: str
    session: str
    run: str
    loader: Any

    @property
    def size(self):
        return self.smax - self.smin

    def run_meta(self) -> Path:
        run_data: Run = self.loader.data[self.subject][self.session][self.run]
        return {
            "ref": run_data.files.tsv,
            "date_time": run_data.datetime_start_og,
            "duration": run_data.duration
        }

    def label(self):
        run_data: Run = self.loader.data[self.subject][self.session][self.run]
        y = run_data.y[self.smin : self.smax]
        y = 1 if y.any() else 0
        return y

    def time(self):
        run_data: Run = self.loader.data[self.subject][self.session][self.run]
        t = run_data.t[self.smin]
        return t

    def data(
        self,
        usage: Literal["all", "train", "test"] = "all"
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Gets data from a segment.

        Args:
            usage (all, train, test, optional): Defaults to "all".

        Returns:
            Tuple[ndarray, ndarray, ndarray]:
            - x (1, C, L): signal data
            - y (L,): label at each sample
            - t (L,): time (secs) since start of data at each sample
        """
        run_data: Run = self.loader.data[self.subject][self.session][self.run]
        win_size = self.loader.win_size
        if not win_size:
            raise ValueError("[segment] loader win_size is missing")

        if usage == "all":
            # Returns all of the segment data
            smin = self.smin
            smax = self.smax

        elif usage == "test":
            # Returns the first window of the segment
            smin = self.smin
            smax = smin + win_size

        elif usage == "train":
            # Returns a random window from the segment
            from random import randint
            smin = randint(self.smin, self.smax - win_size)
            smax = smin + win_size

        else:
            raise NotImplementedError(f"[segment] unknown usage '{usage}'")

        x = run_data.x[..., smin : smax]
        y = run_data.y[smin : smax]
        t = run_data.t[smin : smax]
        return (x, y, t) # Tuple[(N, C, L), (L,), (L,)]

    def plot(self, save_path: Path = None, **kwargs):
        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        x, y, t = self.data(**kwargs)
        N, C, L = x.shape
        fig = plt.figure(figsize=(14, (C * 0.6)))
        gs = GridSpec(nrows=C, ncols=1)
        axes = []

        for c in range(C):
            ax = fig.add_subplot(gs[c, 0], sharex=axes[0] if c > 0 else None)
            ax.plot(t, x[0, c], color="black")
            axes.append(ax)

        plt.suptitle(" ".join([self.subject, self.session, self.run]))

        if save_path:
            plt.savefig(save_path)

        else:
            plt.show(block=True)

        return


@dataclass
class SegmentJob:
    fn: Callable
    win_size: int
    stride: int
    train: bool = True
    dev: bool = True
    test: bool = True


@dataclass
class RunFiles:
    edf: str
    json: str
    tsv: str


@dataclass
class Run:
    subject: str
    session: str
    run: str
    datetime_start_og: datetime = None
    datetime_start: datetime = None
    duration: float = None
    files: RunFiles = None
    x: ndarray | Tensor = None   # (N, C, L)
    y: ndarray | Tensor = None   # (L,)
    t: ndarray | Tensor = None   # (L,)
    _is_tensor: bool = False
    segments_train: List[Segment] = None
    segments_dev: List[Segment] = None
    segments_test: List[Segment] = None


@dataclass
class Session:
    subject: str
    session: str
    datetime_start: datetime = None
    sfreq: float = None
    powerline_freq: float = None
    runs: OrderedDict[str, Run] = None

    def __getitem__(self, index):
        if self.runs is not None:
            return self.runs[index]
        raise IndexError("Runs dict is not initialised")

    def values(self):
        return self.runs.values()

    def keys(self):
        return self.runs.keys()

    def items(self):
        return self.runs.items()


@dataclass
class Subject:
    subject: str
    age: int = None
    sex: str = None
    datetime_start: datetime = None
    sessions: OrderedDict[str, Session] = None

    def __getitem__(self, index):
        if self.sessions is not None:
            return self.sessions[index]
        raise IndexError("Sessions dict is not initialised")

    def values(self):
        return self.sessions.values()

    def keys(self):
        return self.sessions.keys()

    def items(self):
        return self.sessions.items()


@dataclass
class BIDSTree:
    path: Path
    subjects: OrderedDict[str, Subject]
    sfreq: float = None

    def __getitem__(self, index):
        if self.subjects is not None:
            return self.subjects[index]
        raise IndexError("Subjects dict is not initialised")

    def values(self):
        return self.subjects.values()

    def keys(self):
        return self.subjects.keys()

    def items(self):
        return self.subjects.items()

    def apply(self,
        fn: Callable,
        carry: Any = None,
        status: Status = None,
        **kwargs
    ):
        for sub_i, (sub, sub_data) in enumerate(self.items()):
            for ses_i, (ses, ses_data) in enumerate(sub_data.items()):
                for run_i, (run, run_data) in enumerate(ses_data.items()):
                    if status:
                        status.update(" ".join([sub, ses, run]))
                    args = dict(
                        tree=self,
                        sub_i=sub_i, sub=sub, sub_data=sub_data,
                        ses_i=ses_i, ses=ses, ses_data=ses_data,
                        run_i=run_i, run=run, run_data=run_data
                    )
                    carry = fn(args, carry, **kwargs)
        return carry



from pathlib import Path
from typing import List, Dict
import re

from epilepsy_bids_loader import Run, Subject
from epilepsy_bids_loader import BIDSLoader
from epilepsy_bids_loader import CVFold, CrossValidation
from epilepsy_bids_loader import Status, read_yaml


class ManifestBIDSLoader(BIDSLoader):

    def __init__(
        self,
        data_path: str | Path,
        manifest_path: str | Path,
        enforce_chronology: bool,
        allowed_subjects: List[str] = None,
    ) -> None:
        super().__init__(
            data_path=data_path,
            enforce_chronology=enforce_chronology,
            allowed_subjects=allowed_subjects
        )
        self.load_and_check_manifest(manifest_path)
        return

    def load_and_check_manifest(self, manifest_path: Path):

        manifest = read_yaml(manifest_path)

        assert "subject_specific" in manifest.keys()
        assert "subject_independent" in manifest.keys()

        sub_spec = manifest["subject_specific"]
        for sub, cv_folds in sub_spec.items():
            assert re.match(r"^sub-\d{2}$", sub)

            for i, (fold, datasets) in enumerate(cv_folds.items()):
                assert fold == f"fold-{i:02d}"

                for ds_type in ["train", "dev", "test"]:
                    assert ds_type in datasets.keys()

                    for ses, runs in datasets[ds_type].items():
                        assert re.match(r"^ses-\d{2}$", ses)

                        for run in runs:
                            assert re.match(r"^run-\d{2}$", run)

        self.sub_spec: Dict[str, Dict[str, Dict[str, Dict[str, List[str]]]]] = \
            sub_spec

        sub_ind = manifest["subject_independent"]
        for i, (fold, datasets) in enumerate(sub_ind.items()):
            assert fold == f"fold-{i:02d}"

            for ds_type in ["train", "dev", "test"]:
                assert ds_type in datasets.keys()

                for sub in datasets[ds_type]:
                    assert re.match(r"^sub-\d{2}$", sub)

        self.sub_ind: Dict[str, Dict[str, List[str]]] = sub_ind

        return

    def _get_sub_spec_folds(self, sub: str) -> CrossValidation:
        """
        Time-series cross validation (TSCV) for subject specific models
        based on specified manifest.

        Args:
            sub (str): Subject identifier.

        Returns:
            CrossValidation: CV object.
        """
        sub_data: Subject = self.data.subjects[sub]
        manifest = self.sub_spec[sub]

        runs: Dict[str, Dict[str, Run]] = {
            ses: {
                run: run_data
                for run, run_data in ses_data.items()
            }
            for ses, ses_data in sub_data.items()
        }
        status = Status("[loader] subject specific CV ...")
        cv = CrossValidation(method="subject_specific")

        # Create folds from manifest
        for fold in range(len(manifest)):
            status.update(f"fold {fold}/{len(runs) - 1}")
            fold_manifest = manifest[f"fold-{fold:02d}"]
            cv.append(CVFold(
                train=[
                    seg
                    for _ses, _runs in fold_manifest["train"].items()
                    for _run in _runs
                    for seg in runs[_ses][_run].segments_train
                ],
                dev=[
                    seg
                    for _ses, _runs in fold_manifest["dev"].items()
                    for _run in _runs
                    for seg in runs[_ses][_run].segments_dev
                ],
                test=[
                    seg
                    for _ses, _runs in fold_manifest["test"].items()
                    for _run in _runs
                    for seg in runs[_ses][_run].segments_test
                ]
            ))

        status.done()
        return cv

    def _get_sub_indp_folds(self, **kwargs) -> CrossValidation:
        """
        Leave-one-out (subject) cross validation for subject independent models
        based on specified manifest.

        Returns:
            CrossValidation: CV object.
        """
        subs: Dict[str, Subject] = {
            sub: sub_data
            for sub, sub_data in self.data.items()
        }
        manifest = self.sub_ind
        status = Status("[loader] subject independent CV ...")
        cv = CrossValidation(method="subject_independent")
        for fold in range(len(manifest)):
            status.update(f"fold {fold}/{len(subs)}")
            fold_manifest = manifest[f"fold-{fold:02d}"]
            cv.append(CVFold(
                train=[
                    seg
                    for _sub in fold_manifest["train"]
                    for _ses_data in subs[_sub].values()
                    for _run_data in _ses_data.values()
                    for seg in _run_data.segments_train
                ],
                dev=[
                    seg
                    for _sub in fold_manifest["dev"]
                    for _ses_data in subs[_sub].values()
                    for _run_data in _ses_data.values()
                    for seg in _run_data.segments_dev
                ],
                test=[
                    seg
                    for _sub in fold_manifest["test"]
                    for _ses_data in subs[_sub].values()
                    for _run_data in _ses_data.values()
                    for seg in _run_data.segments_test
                ]
            ))
        status.done()
        return cv



from .structs import Segment, SegmentJob
from .structs import Run, RunFiles, Session, Subject
from .structs import BIDSTree
from .utils import Status
from .utils import ZScoreScaler
from .common import batch_segments, read_json, read_yaml
from .cross_validation import CVFold, TestBatch, CrossValidation
from .bids_loader import BIDSLoader
from .manifest_loader import ManifestBIDSLoader
from .segment import get_segment_stats, fixed_segmentation


import numpy as np
from typing import Any, List

from epilepsy_bids_loader import Run, Segment, BIDSLoader


def get_segment_stats(
    args: dict,
    carry: List[float],
    **kwargs
):
    run_data: Run = args["run_data"]
    seg_sizes = [
        seg.size
        for seg in run_data.segments_test
    ]
    return carry + seg_sizes


def fixed_segmentation(
    args: dict,
    carry: Any,
    loader: BIDSLoader,
    win_size: int,
    stride: int,
    train: bool = True,
    dev: bool = True,
    test: bool = True,
    **kwargs
):
    sub = args["sub"]
    ses = args["ses"]
    run = args["run"]
    run_data: Run = args["run_data"]
    N, C, L = run_data.x.shape
    segments = [
        Segment(
            smin=i,
            smax=i + win_size,
            subject=sub,
            session=ses,
            run=run,
            loader=loader
        )
        for i in np.arange(0, (L - win_size), stride)
    ]

    if train:
        run_data.segments_train = segments

    if dev:
        run_data.segments_dev = segments

    if test:
        run_data.segments_test = segments

    return



import torch
from torch import Tensor
from dataclasses import dataclass, field
from pathlib import Path
from datetime import datetime
from typing import Tuple, List, Dict, Literal, Generator

from epilepsy_bids_loader import Segment
from epilepsy_bids_loader import Status, ZScoreScaler
from epilepsy_bids_loader import batch_segments


@dataclass
class CVFold:
    train: List[Segment]
    dev: List[Segment]
    test: List[Segment]


@dataclass
class TestBatch:
    run: str
    ref: Path
    date_time: datetime
    duration: float
    segments: List[Segment] = field(default_factory=list)
    batches: Generator[Tuple[Tensor, Tensor, Tensor], None, None] = None

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        return iter(self.batches)

    def append_segment(self, seg: Segment):
        self.segments.append(seg)
        return

    def create_batches(self, batch_size: int):
        self.batches = batch_segments(
            segments=self.segments,
            batch_size=batch_size,
            usage="test"
        )
        return


@dataclass
class CrossValidation:
    method: Literal["subject_specific", "subject_independent"]
    folds: List[CVFold] = field(default_factory=list)

    def __len__(self):
        return len(self.folds)

    def append(self, fold: CVFold):
        self.folds.append(fold)
        return

    def _train_batch_default(
        self,
        fold: int,
        batch_size: int,
        **kwargs
    ) -> Generator[Tuple[Tensor, Tensor, Tensor], None, None]:
        """
        Standard training batching process
        - Use all bckg, repeat sz, randomise
        """
        from random import sample
        segments: List[Segment] = self.folds[fold].train

        # Balanced selection of sz and bckg segments
        sz_segs = [
            seg for seg in segments
            if seg.label() == 1
        ]
        bckg_segs = [
            seg for seg in segments
            if seg.label() == 0
        ]
        segments = (
            bckg_segs
            + sz_segs * (len(bckg_segs) // len(sz_segs))
        )

        # Randomise
        segments = sample(segments, k=len(segments))

        batches = batch_segments(
            segments=segments,
            batch_size=batch_size,
            usage="train"
        )
        return batches

    def train_batch(
        self,
        fold: int,
        method: Literal["default"],
        batch_size: int,
        **kwargs
    ) -> Generator[Tuple[Tensor, Tensor, Tensor], None, None]:
        _batch_fn = {
            "default": self._train_batch_default,
            # TODO Extend options here as needed
        }[method]
        return _batch_fn(
            fold=fold,
            batch_size=batch_size,
            **kwargs
        )

    def test_batch_by_run(
        self,
        fold: int,
        batch_size: int,
    ) -> Dict[Tuple[str, str, str], TestBatch]:
        """
        Generates batches of test samples grouped by run.

        Args:
            fold (int): The cross validation fold number.
            batch_size (int): Batch size

        Returns:
            Dict[Tuple[str, str, str], TestBatch]: Keys are (sub, ses, run)
        """
        segments: List[Segment] = self.folds[fold].test

        # Organise segments by run
        batches: Dict[str, TestBatch] = {}
        for seg in segments:
            sub = seg.subject
            ses = seg.session
            run = seg.run
            key = (sub, ses, run)
            if key not in batches:
                run_meta = seg.run_meta()
                batches[key] = TestBatch(
                    run=run,
                    ref=run_meta["ref"],
                    date_time=run_meta["date_time"],
                    duration=run_meta["duration"]
                )
            batches[(sub, ses, run)].append_segment(seg)

        # Create batched data by run
        for test_batch in batches.values():
            test_batch.create_batches(batch_size=batch_size)
        return batches

    def fit_scaler(
        self,
        fold: int = 0
    ) -> ZScoreScaler:
        status = Status("[loader] fitting z-score scaler on training data")
        x = torch.concatenate(
            [
                seg.data("all")[0] # (1, C, L)
                for seg in self.folds[fold].train
            ],
            axis=2
        ) # (1, C, L+)
        scaler = ZScoreScaler()
        scaler.fit(x)
        status.done()
        return scaler